@@ -2,6 +2,7 @@ import asyncio
 from contextlib import asynccontextmanager
 from dataclasses import asdict, dataclass, field
 from decimal import Decimal
+from enum import Enum, unique
 from typing import Any, AsyncIterator, Dict, List, Mapping, Optional, Sequence
 
 import aiohttp
@@ -33,6 +34,19 @@ class User:
     clusters: List[Cluster] = field(default_factory=list)
 
 
+@unique
+class Action(str, Enum):
+    READ = "read"
+    WRITE = "write"
+    MANAGE = "manage"
+
+    def __str__(self) -> str:
+        return self.value
+
+    def __repr__(self) -> str:
+        return self.__str__().__repr__()
+
+
 @dataclass(frozen=True)
 class Permission:
     uri: str
@@ -51,6 +65,15 @@ class Permission:
         return check_action_allowed(self.action, "write")
 
 
+@dataclass(frozen=True)
+class Role:
+    name: str
+    permissions: Sequence[Permission]
+
+    def to_role_permission(self, *, action: Action = Action.READ) -> Permission:
+        return Permission(f"role://{self.name}", action)
+
+
 @dataclass
 class ClientAccessSubTreeView:
     action: str
@@ -97,6 +120,7 @@ class AuthClient:
         token: str,
         trace_configs: Optional[List[aiohttp.TraceConfig]] = None,
     ) -> None:
+        # Empty URL and token is used for secure-less mode
         self._token = token
         headers = self._generate_headers(token)
         self._client = aiohttp.ClientSession(
@@ -117,6 +141,7 @@ class AuthClient:
         return headers
 
     def _make_url(self, path: str) -> URL:
+        assert self._url
         if path.startswith("/"):
             path = path[1:]
         return self._url / path
@@ -158,11 +183,15 @@ class AuthClient:
             resp.release()
 
     async def ping(self) -> None:
+        if not self._url:
+            return
         async with self._request("GET", "/api/v1/ping") as resp:
             txt = await resp.text()
             assert txt == "Pong"
 
     async def secured_ping(self, token: Optional[str] = None) -> None:
+        if not self._url:
+            return
         path = "/api/v1/secured-ping"
         headers = self._generate_headers(token)
         async with self._request("GET", path, headers=headers) as resp:
@@ -192,6 +221,8 @@ class AuthClient:
         return f"/api/v1/users/{name}"
 
     async def update_user(self, user: User, token: Optional[str] = None) -> None:
+        if not self._url:
+            return
         path = self._get_user_path(user.name)
         headers = self._generate_headers(token)
         payload = self._serialize_user(user)
@@ -199,6 +230,20 @@ class AuthClient:
             pass  # use context manager to release response earlier
 
     async def get_user(self, name: str, token: Optional[str] = None) -> User:
+        if not self._url:
+            return User(
+                name="user",
+                email="email@example.com",
+                clusters=[
+                    Cluster(
+                        name="default",
+                        quota=Quota(
+                            credits=None,
+                            total_running_jobs=None,
+                        ),
+                    )
+                ],
+            )
         path = self._get_user_path(name)
         headers = self._generate_headers(token)
         async with self._request("GET", path, headers=headers) as resp:
@@ -226,6 +271,8 @@ class AuthClient:
     async def check_user_permissions(
         self, name: str, permissions: Sequence[Permission], token: Optional[str] = None
     ) -> bool:
+        if not self._url:
+            return True
         missing = await self.get_missing_permissions(name, permissions, token)
         return not missing
 
@@ -233,6 +280,8 @@ class AuthClient:
         self, name: str, permissions: Sequence[Permission], token: Optional[str] = None
     ) -> Sequence[Permission]:
         assert permissions, "No permissions passed"
+        if not self._url:
+            return []
         path = self._get_user_path(name) + "/permissions/check"
         headers = self._generate_headers(token)
         payload: List[Dict[str, Any]] = [asdict(p) for p in permissions]
@@ -254,6 +303,11 @@ class AuthClient:
     async def get_permissions_tree(
         self, name: str, resource: str, depth: Optional[int] = None
     ) -> ClientSubTreeViewRoot:
+        if not self._url:
+            return ClientSubTreeViewRoot(
+                path="/default/user",
+                sub_tree=ClientAccessSubTreeView(action="manage", children={}),
+            )
         url = self._get_user_path(name) + "/permissions/tree"
         req_params: Dict[str, Any] = {"uri": resource}
         if depth is not None:
@@ -266,6 +320,8 @@ class AuthClient:
     async def grant_user_permissions(
         self, name: str, permissions: Sequence[Permission], token: Optional[str] = None
     ) -> None:
+        if not self._url:
+            return
         path = self._get_user_path(name) + "/permissions"
         headers = self._generate_headers(token)
         payload: List[Dict[str, str]] = [asdict(p) for p in permissions]
@@ -276,6 +332,8 @@ class AuthClient:
     async def revoke_user_permissions(
         self, name: str, resources_uris: Sequence[str], token: Optional[str] = None
     ) -> None:
+        if not self._url:
+            return
         path = self._get_user_path(name) + "/permissions"
         headers = self._generate_headers(token)
         params = MultiDict(("uri", uri) for uri in resources_uris)
@@ -291,6 +349,8 @@ class AuthClient:
         new_token_uri: Optional[str] = None,
         token: Optional[str] = None,
     ) -> str:
+        if not self._url:
+            return ""
         path = self._get_user_path(name) + "/token"
         headers = self._generate_headers(token)
         if new_token_uri:
@@ -301,6 +361,116 @@ class AuthClient:
             payload = await resp.json()
             return payload["access_token"]
 
+    async def update_role(self, role: User) -> None:
+        if not self._url:
+            return
+        try:
+            await self.update_user(role)
+        except aiohttp.ClientResponseError as exc:
+            if exc.status != 404:
+                raise
+
+            await self.add_user(role)
+
+    async def grant_role_permissions(
+        self,
+        role_name: str,
+        permissions: Sequence[Permission],
+        ignore_existing_role: bool = True,
+    ) -> None:
+        assert permissions
+        if not self._url:
+            return
+        try:
+            await self.grant_user_permissions(role_name, permissions)
+        except aiohttp.ClientResponseError as exc:
+            if exc.status != 404:
+                raise
+
+            await self.add_user(User(name=role_name))
+            await self.grant_user_permissions(role_name, permissions)
+
+    async def set_role_permissions(
+        self,
+        role_name: str,
+        permissions: Sequence[Permission],
+        ignore_existing_role: bool = True,
+    ) -> None:
+        assert permissions
+        if not self._url:
+            return
+        try:
+            await self._set_user_permissions(role_name, permissions)
+        except aiohttp.ClientResponseError as exc:
+            if exc.status != 404:
+                raise
+
+            await self.add_user(User(name=role_name))
+            await self._set_user_permissions(role_name, permissions)
+
+    async def _set_user_permissions(
+        self, name: str, permissions: Sequence[Permission]
+    ) -> None:
+        path = self._get_user_path(name) + "/permissions"
+        payload: List[Dict[str, str]] = [asdict(p) for p in permissions]
+        async with self._request("PUT", path, json=payload) as resp:
+            status = resp.status
+            assert (
+                status == aiohttp.web.HTTPNoContent.status_code
+            ), f"unexpected response: {status}"
+
+    async def revoke_role_permissions(
+        self, role_name: str, permissions: Sequence[Permission]
+    ) -> None:
+        assert permissions
+        if not self._url:
+            return
+        for perm in permissions:
+            try:
+                await self.revoke_user_permissions(role_name, [perm.uri])
+            except aiohttp.ClientResponseError as exc:
+                if exc.status == 400 and exc.message == "Operation has no effect":
+                    pass
+                else:
+                    raise
+
+    async def get_permissions(
+        self, uname: str, *, expand_roles: bool = True
+    ) -> Sequence[Permission]:
+        if not self._url:
+            raise NotImplementedError("The method is not supported by Single-user mode")
+        url = self._get_user_path(uname) + "/permissions"
+        params = {}
+        if not expand_roles:
+            params["expand_roles"] = "false"
+        async with self._request("GET", url, params=params) as resp:
+            payload = await resp.json()
+            return [Permission(item["uri"], item["action"]) for item in payload]
+
+    async def add_role(self, uname: str) -> None:
+        if not self._url:
+            return
+        await self.add_user(User(name=uname))
+
+    async def remove_role(self, uname: str) -> None:
+        if not self._url:
+            return
+        url = self._get_user_path(uname)
+        try:
+            async with self._request("DELETE", url):
+                pass
+        except aiohttp.ClientResponseError as exc:
+            if exc.status != 404:
+                raise
+
+    async def delete_user(self, name: str, token: Optional[str] = None) -> None:
+        if not self._url:
+            return
+        path = self._get_user_path(name)
+        headers = self._generate_headers(token)
+        async with self._request("DELETE", path, headers=headers):
+            pass  # use context manager to release response earlier
+
 
 async def _raise_for_status(resp: aiohttp.ClientResponse) -> None:
     if 400 <= resp.status:
